@@ -2,6 +2,7 @@ using System;
 using System.Collections.Generic;
 using System.Text;
 using Mono.Unix;
+using Gtk;
 
 namespace Tomboy
 {
@@ -25,7 +26,8 @@ namespace Tomboy
 		Gtk.HPaned hpaned;
 		Gtk.VBox content_vbox;
 		Gtk.TreeViewColumn matches_column;
-
+		Gtk.HBox no_matches_box;	//to display a message with buttons when no results are found.
+		
 		Notebooks.NotebooksTreeView notebooksTree;
 
 		// Use the following like a Set
@@ -206,7 +208,30 @@ namespace Tomboy
 			// until the note's QueueSave () kicks in.
 			Notebooks.NotebookManager.NoteAddedToNotebook += OnNoteAddedToNotebook;
 			Notebooks.NotebookManager.NoteRemovedFromNotebook += OnNoteRemovedFromNotebook;
+			
+			// Set the focus chain for the top-most containers Bug #512175
+			Gtk.Widget[] vbox_focus = new Gtk.Widget[2];
+			vbox_focus[0] = hbox;
+			vbox_focus[1] = hpaned;
+			vbox.FocusChain = vbox_focus;
 
+			// Set focus chain for sub widgits of first top-most container
+			Gtk.Widget[] table_focus = new Gtk.Widget[2];
+			table_focus[0] = find_combo;
+			table_focus[1] = matches_window;
+			hbox.FocusChain = table_focus;
+			
+			// set focus chain for sub widgits of seconf top-most container
+			Gtk.Widget[] hpaned_focus = new Gtk.Widget[2];
+			hpaned_focus[0] = matches_window;
+			hpaned_focus[1] = notebooksPane;
+			hpaned.FocusChain = hpaned_focus;
+			
+			// get back to the beginning of the focus chain
+			Gtk.Widget[] scroll_right = new Gtk.Widget[1];
+			scroll_right[0] = tree;
+			matches_window.FocusChain = scroll_right;
+			
 			Tomboy.ExitingEvent += OnExitingEvent;
 		}
 
@@ -317,9 +342,12 @@ namespace Tomboy
 			tree.Selection.Mode = Gtk.SelectionMode.Multiple;
 			tree.Selection.Changed += OnSelectionChanged;
 			tree.ButtonPressEvent += OnTreeViewButtonPressed;
+			tree.KeyPressEvent += OnTreeViewKeyPressed;
 			tree.MotionNotifyEvent += OnTreeViewMotionNotify;
 			tree.ButtonReleaseEvent += OnTreeViewButtonReleased;
 			tree.DragDataGet += OnTreeViewDragDataGet;
+			tree.FocusInEvent += OnTreeViewFocused;
+			tree.FocusOutEvent += OnTreeViewFocusedOut;
 
 			tree.EnableModelDragSource (Gdk.ModifierType.Button1Mask | Gdk.ModifierType.Button3Mask,
 						    targets,
@@ -465,14 +493,22 @@ namespace Tomboy
 
 			IDictionary<Note,int> results =
 				search.SearchNotes(text, false, selected_notebook);
-			foreach (Note note in results.Keys){
-				current_matches.Add(note.Uri, results[note]);
+			
+			// if no results found in current notebook ask user whether
+			// to search in all notebooks
+			if (results.Count == 0 && selected_notebook != null) {
+				NoMatchesFoundAction ();
 			}
+			else {
+				foreach (Note note in results.Keys){
+					current_matches.Add (note.Uri, results[note]);
+				}
 
-			AddMatchesColumn ();
-			store_filter.Refilter ();
-			tree.ScrollToPoint (0, 0);
-			UpdateMatchNoteCount (current_matches.Count);
+				AddMatchesColumn ();
+				store_filter.Refilter ();
+				tree.ScrollToPoint (0, 0);
+				UpdateMatchNoteCount (current_matches.Count);
+			}
 		}
 
 		void AddMatchesColumn ()
@@ -530,7 +566,10 @@ namespace Tomboy
 			if (note != null) {
 				int match_count;
 				if (current_matches.TryGetValue (note.Uri, out match_count)) {
-					if (match_count > 0) {
+					if (match_count == int.MaxValue) {
+						match_str = string.Format (
+								    Catalog.GetString ("Title match"));
+					} else if (match_count > 0) {
 						match_str = string.Format (
 								    Catalog.GetPluralString ("{0} match",
 											     "{0} matches",
@@ -564,6 +603,51 @@ namespace Tomboy
 			status_bar.Pop (0);
 			status_bar.Push (0, status);
 		}
+		
+		// called when no search results are found in the selected notebook
+		void NoMatchesFoundAction ()
+		{
+			hpaned.Remove (matches_window);
+			String message = Catalog.GetString ("No results found " +
+				"in the selected notebook.\nClick here to " +
+				"search across all notes.");
+			Gtk.LinkButton link_button = new Gtk.LinkButton ("", message);
+			Gtk.LinkButton.SetUriHook(ShowAllSearchResults);
+			link_button.TooltipText = Catalog.GetString 
+				("Click here to search across all notebooks");
+			link_button.Show();
+			Gtk.Table no_matches_found_table = new Gtk.Table (1, 3, false);
+			no_matches_found_table.Attach (link_button, 1, 2, 0, 1,
+			                               Gtk.AttachOptions.Fill | Gtk.AttachOptions.Shrink,
+			                 Gtk.AttachOptions.Shrink,
+			                0, 0
+			              );
+			
+			no_matches_found_table.ColumnSpacing = 4;
+			no_matches_found_table.ShowAll ();
+			no_matches_box = new HBox (false, 0);
+			no_matches_box.PackStart (no_matches_found_table, true, true, 0);
+			no_matches_box.Show ();
+			hpaned.Add2 (no_matches_box);
+		}
+		
+		void RestoreMatchesWindow ()
+		{
+			if (no_matches_box != null) {
+				hpaned.Remove (no_matches_box);
+				hpaned.Add2 (matches_window);
+				no_matches_box = null;
+				RestorePosition();
+			}	
+		}
+		
+		private void ShowAllSearchResults (Gtk.LinkButton button, String param)
+		{
+			TreeIter iter;
+			notebooksTree.Model.GetIterFirst (out iter);
+			notebooksTree.Selection.SelectIter (iter);
+		}		
+		
 
 		/// <summary>
 		/// Filter out notes based on the current search string
@@ -652,17 +736,20 @@ namespace Tomboy
 
 		void OnNotesChanged (object sender, Note changed)
 		{
+			RestoreMatchesWindow ();
 			UpdateResults ();
 		}
 
 		void OnNoteRenamed (Note note, string old_title)
 		{
+			RestoreMatchesWindow ();
 			UpdateResults ();
 		}
 
 		void OnNoteSaved (Note note)
 		{
 			var rect = tree.VisibleRect;
+			RestoreMatchesWindow ();
 			UpdateResults ();
 			tree.ScrollToPoint (rect.X, rect.Y);
 		}
@@ -724,7 +811,7 @@ namespace Tomboy
 			if (selected_notes == null || selected_notes.Count == 0) {
 				Tomboy.ActionManager ["OpenNoteAction"].Sensitive = false;
 				Tomboy.ActionManager ["DeleteNoteAction"].Sensitive = false;
-			} else if (selected_notes.Count == 1) {
+			} else if (selected_notes.Count > 0) {
 				Tomboy.ActionManager ["OpenNoteAction"].Sensitive = true;
 				Tomboy.ActionManager ["DeleteNoteAction"].Sensitive = true;
 			} else {
@@ -808,6 +895,30 @@ namespace Tomboy
 			}
 		}
 
+		[GLib.ConnectBefore]
+		void OnTreeViewKeyPressed (object sender, Gtk.KeyPressEventArgs args)
+		{
+			switch (args.Event.Key) {
+			case Gdk.Key.Menu:
+				// Pop up the context menu if a note is selected
+				List<Note> selected_notes = GetSelectedNotes ();
+				if (selected_notes != null && selected_notes.Count > 0) {
+						Gtk.Menu menu = Tomboy.ActionManager.GetWidget (
+						"/MainWindowContextMenu") as Gtk.Menu;
+					PopupContextMenuAtLocation (menu, 0, 0);
+					args.RetVal = true;
+				}
+
+				break;
+			case Gdk.Key.Return:
+			case Gdk.Key.KP_Enter:
+				// Open all selected notes
+				OnOpenNote (this, args);
+				args.RetVal = true;
+				break;
+			}
+		}
+
 		[GLib.ConnectBefore]
 		void OnTreeViewMotionNotify (object sender, Gtk.MotionNotifyEventArgs args)
 		{
@@ -844,6 +955,20 @@ namespace Tomboy
 				tree.Selection.SelectPath (path);
 			}
 		}
+		
+		// called when the user moves the focus into the notes TreeView
+		void OnTreeViewFocused (object sender, EventArgs args)
+		{
+			// enable the Delete Note option in the menu bar 
+			Tomboy.ActionManager ["DeleteNoteAction"].Sensitive = true;
+		}
+		
+		// called when the focus moves out of the notes TreeView
+		void OnTreeViewFocusedOut (object sender, EventArgs args)
+		{
+			// Disable the Delete Note option in the menu bar (bug #647462)
+			Tomboy.ActionManager ["DeleteNoteAction"].Sensitive = false;
+		}
 
 		void PopupContextMenuAtLocation (Gtk.Menu menu, int x, int y)
 		{
@@ -874,30 +999,36 @@ namespace Tomboy
 		void PositionContextMenu (Gtk.Menu menu,
 					  out int x, out int y, out bool push_in)
 		{
-			Gtk.TreeIter iter;
+			int pos_x;
+			int pos_y;
 			Gtk.TreePath path;
-			Gtk.TreeSelection selection;
+			Gtk.TreePath [] selected_rows;
+			Gdk.Rectangle cell_rect;
 
 			// Set default "return" values
-			push_in = false; // not used
+			push_in = true;
 			x = 0;
 			y = 0;
 
-			selection = tree.Selection;
-			if (!selection.GetSelected (out iter))
-				return;
+			// Are we currently in the note list?
+			// else, assume we're in the notebook list
+			Gtk.TreeView currentTree = (tree.HasFocus) ? tree : notebooksTree;
 
-			path = store_sort.GetPath (iter);
+			selected_rows = currentTree.Selection.GetSelectedRows ();
+			// Get TreeView's coordinates
+			currentTree.GdkWindow.GetOrigin (out pos_x, out pos_y);
 
-			int pos_x = 0;
-			int pos_y = 0;
+			if (selected_rows.Length > 0) {
+				// Popup near the selection
+				path = selected_rows [0];
+				cell_rect = currentTree.GetCellArea (path, currentTree.Columns [0]);
 
-			GetWidgetScreenPos (tree, ref pos_x, ref pos_y);
-			Gdk.Rectangle cell_rect = tree.GetCellArea (path, tree.Columns [0]);
+				// Add 100 to x, so it isn't too close to the left border
+				x = pos_x + cell_rect.X + 100;
+				// Add 47 to y, so it's right at the bottom of the selected row
+				y = pos_y + cell_rect.Y + 47;
 
-			// Add 100 to x so it's not be at the far left
-			x = pos_x + cell_rect.X + 100;
-			y = pos_y + cell_rect.Y;
+			}
 		}
 
 		// Walk the widget hiearchy to figure out
@@ -965,10 +1096,9 @@ namespace Tomboy
 		void OnOpenNote (object sender, EventArgs args)
 		{
 			List<Note> selected_notes = GetSelectedNotes ();
-			if (selected_notes == null || selected_notes.Count != 1)
-				return;
-
-			selected_notes [0].Window.Present ();
+			if (selected_notes != null)
+				foreach (Note note in selected_notes)
+					note.Window.Present ();
 		}
 
 		void OnDeleteNote (object sender, EventArgs args)
@@ -1030,16 +1160,6 @@ namespace Tomboy
 			case Gdk.Key.Escape:
 				// Allow Escape to close the window
 				OnCloseWindow (this, EventArgs.Empty);
-				break;
-			case Gdk.Key.Menu:
-				// Pop up the context menu if a note is selected
-				List<Note> selected_notes = GetSelectedNotes ();
-				if (selected_notes != null && selected_notes.Count > 0) {
-						Gtk.Menu menu = Tomboy.ActionManager.GetWidget (
-						"/MainWindowContextMenu") as Gtk.Menu;
-				    PopupContextMenuAtLocation (menu, 0, 0);
-				}
-
 				break;
 			}
 		}
@@ -1159,6 +1279,8 @@ namespace Tomboy
 				entry_changed_timeout.Reset (500);
 				clear_search_button.Sensitive = true;
 			}
+			
+			RestoreMatchesWindow ();
 		}
 
 		// Called in after .5 seconds of typing inactivity, or on
@@ -1235,13 +1357,16 @@ namespace Tomboy
 			if (notebook is Notebooks.SpecialNotebook) {
 				// Bold the "Special" Notebooks
 				crt.Markup = string.Format ("<span weight=\"bold\">{0}</span>", notebook.Name);
+				crt.Editable = false;
 			} else {
 				crt.Text = notebook.Name;
+				crt.Editable = true;
 			}
 		}
 
 		private void OnNotebookSelectionChanged (object sender, EventArgs args)
 		{
+			RestoreMatchesWindow ();
 			Notebooks.Notebook notebook = GetSelectedNotebook ();
 			if (notebook == null) {
 				// Clear out the currently selected tags so that no notebook is selected
@@ -1411,11 +1536,13 @@ namespace Tomboy
 
 		private void OnNoteAddedToNotebook (Note note, Notebooks.Notebook notebook)
 		{
+			RestoreMatchesWindow ();
 			UpdateResults ();
 		}
 
 		private void OnNoteRemovedFromNotebook (Note note, Notebooks.Notebook notebook)
 		{
+			RestoreMatchesWindow ();
 			UpdateResults ();
 		}
 
